"""
Views for Djax.
"""
from djax.content import sync_content, sync_record
from django.http import HttpResponse
import base64
from djax.models import AuthToken
from django.conf import settings

def check_auth(view):
    """
    Decorator to check auth token.
    """
    def wrapper(request,*args,**kwargs):
        if hasattr(settings,'DJAX_DISABLE_AUTH') and settings.DJAX_DISABLE_AUTH:
            return view(request,*args,**kwargs)
        else:
            if 'HTTP_AUTHORIZATION' in request.META:
                basic_flag, auth_string = request.META['HTTP_AUTHORIZATION'].split()
                token = base64.b64decode(auth_string)[:-1] # chop trailing colon
                try:
                    auth_token = AuthToken.objects.get(token=token)
                    if auth_token.origin_domain:
                        origin_domain = request.META.get('REMOTE_HOST',None)
                        if not origin_domain == auth_token.origin_domain:
                            return HttpResponse('Not Allowed',status=403)
                
                    return view(request,*args,**kwargs)
                except AuthToken.DoesNotExist:
                    pass
        
            return HttpResponse('Not Allowed',status=403)
    
    return wrapper

def phone_home(request,token=None):
    """
    Initiates a content sync as long as there is no existing sync lock.
    """
    sync_content(token)

@check_auth
def sync_record_view(request):
    """
    Syncs a single record.
    """
    content_type = request.GET.get('content_type',None)
    content_key = request.GET.get('content_key',None)
    if content_type and content_key:
        try:
            if sync_record(content_type,content_key):
                return HttpResponse('Created %s:%s' % (content_type,content_key),status=201)
            else:
                return HttpResponse('Updated %s:%s' % (content_type,content_key))
        except ValueError as ve:
            return HttpResponse(unicode(ve),status=409) # probably a mismatch of content types
    else:
        return HttpResponse('Badly formed request, specify content_type and content_key in request params.',status=409)



"""
Makes an auth token for remote management.
"""
from django.core.management.base import BaseCommand
from optparse import make_option
from djax.models import AuthToken

class Command(BaseCommand):
    """
    Command class.
    """
    option_list = BaseCommand.option_list + (
        make_option('--host',dest='host',default=None),
    )
    
    def handle(request,*args,**options):
        """
        Handler method.
        """
        origin_domain = options['host']
        token = AuthToken.objects.new_token(origin_domain)
        print 'created token',token



"""
Revokes an auth token for remote management.
"""
from django.core.management.base import BaseCommand
from optparse import make_option
from djax.models import AuthToken

class Command(BaseCommand):
    """
    Command class.
    """
    option_list = BaseCommand.option_list + (
        make_option('--token',dest='token',default=None),
    )
    
    def handle(request,*args,**options):
        """
        Handler method.
        """
        try:
            token = AuthToken.objects.get(token=options['token'])
            token.delete()
            print 'Deleted token',options['token']
        except AuthToken.DoesNotExist:
            print 'Cannot find token',options['token'],'in this Djax install.'



"""
For calendar module.
"""
from pax.content import ContentImage
from dateutil import parser

class EventImage(object):
    """
    Local wrapper for event data.
    """
    def __init__(self,data):
        self.content = ContentImage(data['content'])
        self.calendar = data['calendar']
        self.event_type = data['event_type']
        self.start = parser.parse(data['start'])
        self.end = parser.parse(data['end'])
    
    def __getattr__(self,attribute):
        return getattr(self.content,attribute)

class ResourceImage(object):
    """
    Local wrapper for a resource.
    """
    def __init__(self,data):
        self.profile = data['profile']
        self.name = data['name']

class ResourceAvailabilityBlock(object):
    """
    Wraps availability for a resource.
    """
    def __init__(self,profile,start,end):
        self.profile = profile
        self.start = start
        self.end = end

class AggregateAvailabiltyBlock(object):
    """
    Wrapps aggregate availability.
    """
    def __init__(self,start,end):
        self.start = start
        self.end = end

class AvailabilityImage(object):
    """
    Local wrapper for data indicating resource availability.
    """
    def __init__(self,data):
        self.availabilty = [AggregateAvailabilityBlock(avail['start'],avail['end']) for avail in data['availability']]
        self.resources = {}
        for key,value in data['by-resource']:
            self.resources[key] = [ResourceAvailabilityBlock(key,avail['start'],avail['end']) for avail in value]

class CalendarClient(object):
    """
    Client for Calendar API.
    """
    def __init__(self,axilent_connection):
        self.event_resource = axilent_connection.resource_client('axilent.plugins.calendar','event')
        self.calendar_resource = axilent_connection.resource_client('axilent.plugins.calendar','calendar')
        self.resource_resource = axilent_connection.resource_client('axilent.plugins.calendar','resource')
        
        self.api = axilent_connection.http_client('axilent.plugins.calendar')
    
    def get_event(self,calendar,key):
        """
        Gets the specified event.
        """
        data = self.event_resource.get(params={'calendar':calendar,'key':key})
        return EventImage(data)
    
    def create_event(self,
                     calendar,
                     event_type,
                     start,
                     end,
                     content,
                     recurrence_quantity=None,
                     recurrence_unit=None,
                     recurrence_end=None,
                     resources=None):
        """
        Creates a new event.
        """
        response = self.event_resource.post(data={'calendar':calendar,
                                                    'event_type':event_type,
                                                    'start':start.strftime('%Y-%m-%d %H:%M:%S'),
                                                    'end':end.strftime('%Y-%m-%d %H:%M:%S'),
                                                    'content':content,
                                                    'recurrence_quantity':recurrence_quantity,
                                                    'recurrence_unit':recurrence_unit,
                                                    'recurrence_end':recurrence_end,
                                                    'resources':resources})
        
        return response['event-created']
    
    def update_event(self,
                     calendar,
                     key,
                     event_type=None,
                     start=None,
                     end=None,
                     recurrence_quantity=None,
                     recurrence_unit=None,
                     recurrence_end=None,
                     content=None):
        """
        Updates an existing event.
        """
        response = self.event_resource.put(data={'calendar':calendar,
                                                   'key':key,
                                                   'event_type':event_type,
                                                   'start':start.strftime('%Y-%m-%d %H:%M:%S'),
                                                   'end':end.strftime('%Y-%m-%d %H:%M:%S'),
                                                   'recurrence_quantity':recurrence_quantity,
                                                   'recurrence_unit':recurrence_unit,
                                                   'recurrence_end':recurrence_end,
                                                   'content':content})
    
    def delete_event(self,calendar,key):
        """
        Deletes the event.
        """
        self.event_resource.delete(params={'calendar':calendar,'key':key})
    
    def clear_event_recurrence(self,calendar,key):
        """
        Clears recurrence on an event.
        """
        response = self.api.cleareventrecurrence(calendar=calendar,key=key)
        if response['event-recurrence'] == 'cleared':
            return True
        else:
            return False
    
    def list_events(self,calendar,start,end,event_types=None,resources=None,ical=False):
        """
        Gets a list of events for the specified calendar, within the specified date range.
        
        If event types are specified only events of those event types will be returned.
        If resources are specified, only resources where there are bookings matching the
        specified resources will be returned.
        """
        response = self.api.listevents(calendar=calendar,
                                       start=start.strftime('%Y-%m-%d %H:%M:%S'),
                                       end=end.strftime('%Y-%m-%d %H:%M:%S'),
                                       event_types=event_types,
                                       resources=resources,
                                       render_ical=unicode(ical))
        
        return response
    
    def list_calendars(self):
        """
        Lists the calendars.
        """
        return self.calendar_resource.get()
    
    def create_calendar(self,calendar):
        """
        Creates a calendar.
        """
        self.calendar_resource.post(params={'calendar':calendar})
    
    def delete_calendar(self,calendar):
        """
        Deletes the calendar.
        """
        self.calendar_resource.delete(params={'calendar':calendar})
    
    def get_resource(self,profile):
        """
        Gets the specified resource.
        """
        response = self.resource_resource.get(params={'profile':profile})
        return ResourceImage(response)
    
    def create_resource(self,profile,name):
        """
        Creates a resource.
        """
        response = self.resource_resource.post(params={'profile':profile,'name':name})
        return response['resource-created'] # returns the profile key for the created resource
    
    def update_resource(self,profile,name):
        """
        Updates a resource.
        """
        self.resource_resource.put(params={'profile':profile,'name':name})
    
    def delete_resource(self,profile):
        """
        Deletes the resource.
        """
        self.resource_resource.delete(params={'profile':profile})
    
    def list_events_by_date_range(self,calendars,start,end,resources=None,event_types=None):
        """
        Lists the specified events.
        """
        response = self.api.listeventsbydaterange(calendars=calendars,
                                                  start=start,
                                                  end=end,
                                                  resources=resources,
                                                  event_types=event_types)
        
        return [EventImage(event) for event in response]
    
    def list_events_for_month(self,calendars,year,month,resources=None,event_types=None):
        """
        Lists events for the month.
        """
        response = self.api.listeventsformonth(calendars=calendars,
                                               year=year,
                                               month=month,
                                               resources=resources,
                                               event_types=event_types)
        
        return [EventImage(event) for event in response]
    
    def list_events_for_week(self,calendars,year,week,resources=None,event_types=None):
        """
        Lists events for the week.
        """
        response = self.api.listeventsforweek(calendars=calendars,
                                              year=year,
                                              week=week,
                                              resources=resources,
                                              event_types=event_types)
        
        return [EventImage(event) for event in response]
    
    def get_availability_by_date_range(self,calendars,resources,start,end):
        """
        Gets resource availability within the specified date range.
        """
        response = self.api.getavailabilitybydaterange(calendars=calendars,
                                                       resources=resources,
                                                       start=start,
                                                       end=end)
        
        return AvailabilityImage(response)
    
    
    def get_availability_for_month(self,calendars,resources,year,month):
        """
        Gets resource availability for the specified month.
        """
        response = self.api.getavailabilityformonth(calendars=calendars,
                                                    resources=resources,
                                                    year=year,
                                                    month=month)
        
        return AvailabilityImage(response)
    
    def get_availability_for_week(self,calendars,resources,year,week):
        """
        Gets resources availability for the specified week.
        """
        response = self.api.getavailabilityforweek(calendars=calendars,
                                                   resources=resources,
                                                   year=year,
                                                   month=month)
        
        return AvailabilityImage(response)



"""
Content sync for Djax
"""
from djax.registry import content_registry, build_registry
from djax.gateway import content_client, trigger_client
import uuid
import logging
from django.db.models import Manager
from django.contrib.contenttypes.models import ContentType
from pax.util import slugify
import re

log = logging.getLogger('djax')

class ACEContent(object):
    """
    Mixin to provide Axilent content sync services for Django models.
    """
    def get_axilent_content_key(self):
        """
        Gets the axilent content key for this content.
        """
        from djax.models import AxilentContentRecord
        try:
            record = AxilentContentRecord.objects.get_record(self)
            return record.axilent_content_key
        except AxilentContentRecord.DoesNotExist:
            return None
        
    def get_axilent_content_type(self):
        """
        Gets the axilent content type for this content.
        """
        from djax.models import AxilentContentRecord
        try:
            record = AxilentContentRecord.objects.get_record(self)
            return record.axilent_content_type
        except AxilentContentRecord.DoesNotExist:
            return None
    
    def sync_with_axilent(self):
        """
        Synchronizes the local content with the latest from Axilent.
        """
        from djax.models import AxilentContentRecord
        # 1. Check status of content at Axilent project.  Abort if not updated later than ContentRecord.
        record = AxilentContentRecord.objects.get_record(self)
        if record.update_available():
            axilent_content = record.get_update() # 2. get the updated content from Axilent
            record.sync_content(axilent_content) # 3. write to the local model
    
    def to_content_dict(self):
        """
        Renders the local model as a content dictionary.
        """
        try:
            content_dict = {}
            for axilent_field, local_field in self.ACE.field_map.items():
                try:
                    content_dict[axilent_field] = getattr(self,local_field)
                except AttributeError:
                    raise ValueError('The local field %s is not defined in this model.' % local_field)
            
            return content_dict
        except AttributeError:
            raise ValueError('You must define an ACE field map to create a content dict from a local model.')
    
    def push_to_library(self):
        """
        Pushes this model to the Axilent library, assuming library integration is enabled.
        
        Returns a 2-tuple of booleans indicating 1.  If the library was updated and 2. If the
        content item was created on Axilent for the first time.
        """
        from djax.models import AxilentContentRecord
        return AxilentContentRecord.objects.push_to_library(self)
    
    def push_to_graphstack(self):
        """
        Pushes this model to the graphstack associated with the active content client.
        """
        from djax.models import AxilentContentRecord
        return AxilentContentRecord.objects.push_to_graphstack(self)
    
    def archive(self):
        """
        Archives this content on Axilent.  Archived content will be removed from the workflow,
        and will be un-deployed from any Deployment Targets where it has been deployed.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.archive()
    
    def live_delete(self):
        """
        Deletes the deployed version of the content on the active graphstack.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.live_delete()
        record.delete()
    
    def tag(self,tag_term,update_library_index=True):
        """
        Tags this content.  If update_library_index is set to false then the
        tagging will not update the library search index for the content item.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.tag(tag_term,update_library_index=update_library_index)
    
    def detag(self,tag_term):
        """
        Disassociates the content from the specified tag term.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.detag(tag_term)
    
    def live_tag(self,tag_term):
        """
        Tags the deployed version of the content.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.live_tag(tag_term)
    
    def live_detag(self,tag_term):
        """
        De-tags the deployed version of the content.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.live_detag(tag_term)
    
    def reindex_search(self):
        """
        Re-indexes the deployed version of the content for search.
        """
        from djax.models import AxilentContentRecord
        record = AxilentContentRecord.objects.get_record(self)
        record.reindex()
    
    def trigger_affinity(self,profile,environment={},identity={}):
        """
        Sends affinity trigger for this content.
        """
        trigger_client.trigger('affinity',
                               slugify(self.ACE.content_type),
                               profile=profile,
                               variables={'key':self.get_axilent_content_key()},
                               environment=environment,
                               identity=identity)
    
    def trigger_ban(self,profile,environment={},identity={}):
        """
        Sends a ban trigger for this content.
        """
        trigger_client.trigger('ban',
                               slugify(self.ACE.content_type),
                               profile=profile,
                               variables={'key':self.get_axilent_content_key()},
                               environment=environment,
                               identity=identity)

# ====================
# = Field Converters =
# ====================
class DefaultFieldConverter(object):
    def __init__(self,local_field):
        self.field = local_field
    
    def to_local_model(self,ace_content,ace_field_value):
        """
        Associated content link, if found, otherwise just assigns value.
        """
        from djax.models import AxilentContentRecord
        if re.match(r'^[\w\s]+:[A-Fa-f0-9]+$',ace_field_value):
            ctype, ckey = ace_field_value.split(':')
            local_model = AxilentContentRecord.objects.get_or_create_model(axilent_content_type=ctype,axilent_content_key=ckey)
            log.debug('Converted content link %s to local model.' % ace_field_value)
            return local_model
        else:
            return ace_field_value
    
    def to_ace(self,local_model):
        """
        Returns the content link string from the model.
        """
        from djax.models import AxilentContentRecord
        value = getattr(local_model,self.field)
        if isinstance(value,models.Model):
            try:
                record = AxilentContentRecord.objects.get_record(value)
                log.debug('Converted local model to content link %s:%s.' % (record.axilent_content_type,record.axilent_content_key))
                return '%s:%s' % (record.axilent_content_type,record.axilent_content_key)
            except AxilentContentRecord.DoesNotExist:
                return ''
        else:
            return value

class NullableForeignKeyConverter(object):
    """
    Special converter to convert content links to local foreign keys. The default
    converter would normally handle this, but not in the case where there is the
    possibility of a nullable foreign key.  If the model has a nullable foreign
    key, this converter should be specified instead.
    """
    def __init__(self,local_field):
        self.field = local_field
    
    def to_local_model(self,ace_content,ace_field_value):
        """
        Associated content link or None.
        """
        from djax.models import AxilentContentRecord
        if ace_field_value:
            if re.match(r'^[\w\s]+:[A-Fa-f0-9]+$',ace_field_value):
                ctype, ckey = ace_field_value.split(':')
                local_model = AxilentContentRecord.objects.get_or_create_model(axilent_content_type=ctype,axilent_content_key=ckey)
                log.debug('Converted content link %s to local model.' % ace_field_value)
                return local_model
            else:
                raise ValueError('Cannot use the NullableForeignKeyConverter to set non-null values that are not foreign keys. Value %s for field %s.' % (ace_field_value,self.field))
        else:
            return None
    
    def to_ace(self,local_model):
        """
        Returns a content link string, or ''.
        """
        from djax.models import AxilentContentRecord
        value = getattr(local_model,self.field)
        if value:
            if isinstance(value,models.Model):
                try:
                    record = AxilentContentRecord.objects.get_record(value)
                    log.debug('Converted local model to content link %s:%s.' % (record.axilent_content_type,record.axilent_content_key))
                    return '%s:%s' % (record.axilent_content_type,record.axilent_content_key)
                except AxilentContentRecord.DoesNotExist:
                    return ''
            else:
                raise ValueError('Cannot use the NullableForeignKeyConverter to set non-null values that are not content links. Value %s for field %s.' & (unicode(value),self.field))
        else:
            return ''

class M2MFieldConverter(object):
    """
    Field converter that translates m-2-m model relations, from ACE content link lists.
    """
    deferred = True
    
    def __init__(self,local_field):
        self.field = local_field
    
    def to_local_model(self,ace_content,ace_field_value,local_model):
        """
        Converts to local model relations.
        """
        from djax.models import AxilentContentRecord
        log.debug('Converting ace value %s to local model relations.' % ace_field_value)
        added_models = []
        for compound_key in ace_field_value:
            ctype, ckey = compound_key.split(':')
            linked_model = AxilentContentRecord.objects.get_or_create_model(axilent_content_type=ctype,
                                                                           axilent_content_key=ckey)
            getattr(local_model,self.field).add(linked_model)
            added_models.append(linked_model.pk)
            log.debug('Building local model relation from content link %s.' % compound_key)
        
        # get rid of any relations not specified in the content link list
        for stale_link in getattr(local_model,self.field).exclude(pk__in=added_models):
            getattr(local_model,self.field).remove(stale_link)
            log.debug('Removing stale link %s from local model.' % unicode(stale_link))
        
        local_model.save()
    
    def to_ace(self,local_model):
        """
        Converts local relations to content key list.
        """
        from djax.models import AxilentContentRecord
        key_list = []
        for linked_model in getattr(local_model,self.field).all():
            try:
                record = AxilentContentRecord.objects.get_record(linked_model)
                key_list.append('%s:%s' % (record.axilent_content_type,record.axilent_content_key))
            except AxilentContentRecord.DoesNotExist:
                log.warning('Cannot assign local model %s to ACE content link list.  Local model does not exist in ACE.' % unicode(linked_model))
        
        return key_list

class WriteNullFieldConverter(DefaultFieldConverter):
    """
    Either writes the value supplied from ACE, or None to the local model.
    """
    accepts_null = True  


# ======================
# = Content Operations =
# ======================

def sync_record(content_type,content_key):
    """
    Syncs a single content record.
    """
    from djax.models import AxilentContentRecord
    record = None
    try:
        record = AxilentContentRecord.objects.get(axilent_content_type=content_type,
                                                  axilent_content_key=content_key)
        axilent_content = record.get_update()
        if axilent_content:
            log.info('Syncing local model %s:%s with updated content from ACE.' % (record.axilent_content_type,record.axilent_content_key))
            record.sync_content(axilent_content)
            return False # no new record created
    except AxilentContentRecord.DoesNotExist:
        AxilentContentRecord.objects.create_model(content_type,content_key)
        return True # new record created

def sync_content_type(content_type):
    """
    Syncs a specific content type.
    """
    from djax.models import AxilentContentRecord
    
    content_keys = content_client.content_keys(content_type)
    for content_key in content_keys:
        try:
            record = AxilentContentRecord.objects.get(axilent_content_type=content_type,
                                                      axilent_content_key=content_key)
            axilent_content = record.get_update()
            if axilent_content:
                log.debug('Syncing local model %s:%s with updated content from ACE' % (record.axilent_content_type,record.axilent_content_key))
                record.sync_content(axilent_content)
        except AxilentContentRecord.DoesNotExist:
            AxilentContentRecord.objects.create_model(content_type,content_key)

def sync_content(token=None,content_type_to_sync=None):
    """
    Synchronizes the local models with Axilent content.
    """
    from djax.models import ContentSyncLock
    
    if ContentSyncLock.objects.all().exists():
        return False # already sync locked
    
    if not token:
        token = uuid.uuid4().hex
    
    lock = ContentSyncLock.objects.create(token=token)
    
    # ensure content registry loaded
    build_registry()
    
    if content_type_to_sync:
        log.info('Syncing %s.' % content_type_to_sync)
        try:
            content_type = content_registry[content_type_to_sync]
            sync_content_type(content_type_to_sync)
        except KeyError:
            log.error('%s is not in the content registry.' % content_type_to_sync)
    else:
        for content_type in content_registry.keys():
            sync_content_type(content_type)
    
    lock.delete()
    return True # sync occured

# ===================
# = Content Channel =
# ===================
class ContentChannel(object):
    """
    Accesses an ACE content channel.
    """
    def __init__(self,name=None,flavor=None,limit=0):
        self.name = name
        self.flavor = flavor
        self.limit = limit
        self.api = content_client
    
    def _build_params(self,**params):
        p = {}                
        if 'profile' in params:
            p['profile'] = params['profile']
        
        if 'basekey' in params:
            p['basekey'] = params['basekey']
        
        if 'flavor' in params or self.flavor:
            p['flavor'] = params['flavor'] if 'flavor' in params else self.flavor
        
        if 'limit' in params or self.limit:
            p['limit'] = params['limit'] if 'limit' in params else self.limit
        
        return p
    
    def get_content(self,queryset,channel=None,profile=None,basekey=None,flavor=None,limit=0):
        """
        Gets content.  Flavor and limit params will override the defaults.
        Extracts relevant content from the supplied queryset.
        """
        from djax.models import AxilentContentRecord
        
        ctype = ContentType.objects.get_for_model(queryset.model)
        params = self._build_params(profile=profile,basekey=basekey,flavor=flavor,limit=limit)
        
        if not channel and not self.name:
            raise ValueError('Content Channel unspecified.  You must either specify the channel in the call or the constructor.')
        
        channel_slug = slugify(channel or self.channel)
        
        axl_content_type = queryset.model.ACE.content_type
        
        results = self.api.channel(channel_slug,**params)
        
        return_set = []
        
        for content_item in results.items:
            try:
                record = AxilentContentRecord.objects.get(axilent_content_type=content_item.content_type,
                                                          axilent_content_key=content_item.key)
                return_set.append(record.get_local_model())
            except AxilentContentRecord.DoesNotExist:
                log.warn('No local record of %s:%s, referenced by Content Channel %s' % (axl_content_type,content_item.key,self.name))
            
        return return_set
    
    def __call__(self,queryset,channel=None,profile=None,basekey=None,flavor=None,limit=0):
        """
        Function hook - passes through to get_content.
        """
        return self.get_content(queryset,channel=channel,profile=profile,basekey=basekey,flavor=flavor,limit=limit)

# ===========================================
# = Manager class provides Search Interface =
# ===========================================
class ContentManager(Manager):
    """
    A Manager class that provides access to the ACE search interface and content channels.
    """
    def __init__(self,channel=None,flavor=None,limit=0):
        super(ContentManager,self).__init__()
        
        if channel:
            self._channel = ContentChannel(channel=channel,flavor=flavor,limit=limit)
        else:
            self._channel = ContentChannel()
    
    def search(self,query):
        """
        Returns models that correspond to the search results.
        """
        from djax.models import AxilentContentRecord
        return AxilentContentRecord.objects.search(self.model,query)
    
    def channel(self,channel=None,profile=None,basekey=None,flavor=None,limit=0):
        """
        Gets content matching the channel results
        """
        if not self._channel.name and not channel:
            raise ValueError('Content Channel not defined.  You must either specify it as an argument, or pass a default channel to the constructor.')
        
        return self._channel(self.all(),channel=channel,profile=profile,basekey=basekey,flavor=flavor,limit=limit)
        



"""
Client for the library API.
"""

class LibraryClient(object):
    """
    Library API client.
    """
    def __init__(self,axilent_connection):
        self.content_resource = axilent_connection.resource_client('axilent.library','content')
        self.api = axilent_connection.http_client('axilent.library')
    
    def create_content(self,content_type,project,search_index=True,**field_data):
        """
        Creates the content.  Returns the new content item key in the format:
        
        <content-type>:<content-key>
        """
        response = self.content_resource.post(data={'content_type':content_type,
                                                    'project':project,
                                                    'search_index':search_index,
                                                    'content':field_data})
        return response['created_content']
    
    def update_content(self,content_type,project,content_key,search_index=True,reset_workflow=True,**field_data):
        """
        Updates existing content.
        """
        response = self.content_resource.put(data={'content_type':content_type,
                                                   'project':project,
                                                   'key':content_key,
                                                   'search_index':search_index,
                                                   'reset_workflow':reset_workflow,
                                                   'content':field_data})
        return response['updated_content']
    
    def ping(self,project,content_type):
        """
        Tests connection with Axilent.
        """
        return self.api.ping(project=project,content_type=content_type)
    
    def index_content(self,project,content_type,content_key):
        """
        Forces re-indexing of the specified content item.
        """
        response = self.api.indexcontent(content_key=content_key,
                                         project=project,
                                         content_type=content_type)
        return response['indexed']
    
    def tag_content(self,project,content_type,content_key,tag,search_index=True):
        """
        Tags the specified content item.
        """
        response = self.api.tagcontent(project=project,
                                       content_type=content_type,
                                       content_key=content_key,
                                       tag=tag,
                                       search_index=search_index)
        return response['tagged_content']
    
    def detag_content(self,project,content_type,content_key,tag,search_index=True):
        """
        De-tags the specified content item.
        """
        response = self.api.detagcontent(project=project,
                                         content_type=content_type,
                                         content_key=content_key,
                                         tag=tag,
                                         search_index=search_index)
        return response['removed_tag']
    
    def archive_content(self,project,content_type,content_key):
        """
        Archives the content on Axilent.
        """
        response = self.content_resource.delete(params={'content_type':content_type,
                                                        'project':project,
                                                        'key':content_key})
        return response['archived']



"""
Decorators for Djax.  To be applied to views.
"""
from djax.models import ProfileRecord
from djax.content import ACEContent

def affinity_trigger(model_class,id_name):
    """
    Sends an affinity trigger for the identified model with the specified id name.
    The trigger will pull the id from the incoming argument to the view.
    """
    def func_builder(func):
        def view(request,*args,**kwargs):
            model_id = kwargs[id_name]
            model = model_class.objects.get(pk=model_id)
            profile = ProfileRecord.objects.for_user(request.user)
            
            # sanity check
            if not isinstance(model,ACEContent):
                raise ValueError('Model %s is not ACE Content.' % unicode(model))
            
            model.trigger_affinity(profile)
            
            return func(request,*args,**kwargs)
        
        return view
    
    return func_builder

def ban_trigger(model_class,id_name):
    """
    Sends a ban trigger for the model.
    """
    def func_builder(func):
        def view(request,*args,**kwargs):
            model_id = kwargs[id_name]
            model = model_class.objects.get(pk=model_id)
            profile = ProfileRecord.objects.for_user(request.user)
            
            # sanity check
            if not isinstance(model,ACEContent):
                raise ValueError('Model %s is not ACE Content.' % unicode(model))
            
            model.trigger_ban(profile)
            
            return func(request,*args,**kwargs)
        
        return view
    
    return func_builder

def ensure_profile(view):
    """
    Extracts or creates profile.  Will be passed to wrapped view as 'ace_profile'
    keyword argument.
    """
    def wrapper(request,*args,**kwargs):
        record, created = ProfileRecord.objects.for_request(request)
        response = view(request,ace_profile=record.profile,*args,**kwargs)
        if created:
            response.set_cookie('ace-profile',record.profile)
        return response
    
    return wrapper



"""
Models for Djax.
"""
from django.db import models, IntegrityError
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
import logging
from datetime import datetime
from djax.gateway import content_client, library_client, library_project, trigger_client
from djax.registry import content_registry, build_registry
import re
import threading
import uuid

log = logging.getLogger('djax')

class AxilentContentRecordManager(models.Manager):
    """
    Manager class for AxilentContentRecord.
    """
    def __init__(self,*args,**kwargs):
        super(AxilentContentRecordManager,self).__init__(*args,**kwargs)
        self.lock = threading.RLock()
    
    def get_record(self,model):
        """
        Gets the record for the specified model.
        """
        content_type = ContentType.objects.get_for_model(model)
        return self.get(local_content_type=content_type,local_id=model.pk)
    
    def field_map(self,model,axilent_content={}):
        """
        Gets the field map for the model.
        """
        field_map = {}
        excludes = []
        
        # collect mappings
        if hasattr(model,'ACE'):
            if hasattr(model.ACE,'field_map'):
                field_map = model.ACE.field_map
            
            if hasattr(model.ACE,'exclude'):
                excludes = model.ACE.exclude
        
        # Default the field map to the axilent fields
        if not field_map:
            for key in axilent_content.data.keys():
                if not key in excludes:
                    field_map[key] = key
        
        return field_map
    
    def create_model(self,axilent_content_type,axilent_content_key):
        """
        Creates a new model and accompaning content record for the axilent content.
        """
        from djax.content import DefaultFieldConverter
        
        log.debug('Creating new model of content type:%s and key:%s.' % (axilent_content_type,axilent_content_key))
        
        build_registry() # ensure registry is built
        content_data = content_client.get_content(axilent_content_type,axilent_content_key)
        local_model, record = None, None
        try:
            model_class = content_registry[axilent_content_type]
        
            field_map = {}
            try:
                field_map = model_class.ACE.field_map
            except AttributeError:
                for key in content_data.data.keys():
                    field_map[key] = key
            
            # Iterate through the field map and set the local model values from the incoming Axilent content
            fields = {}
            deferred_field_converters = []
            for axilent_field, model_field in field_map.items():
                if hasattr(content_data,axilent_field):
                    try:            
                        if hasattr(model_field,'field'):
                            log.debug('%s using field converter %s.' % (axilent_field,unicode(model_field)))
                            # this is a field converter
                            # sanity check
                            if not hasattr(model_field,'to_ace') or not hasattr(model_field,'to_local_model'):
                                raise ValueError('You must define the methods to_ace and to_local_model for field converter for ace field %s.' % axilent_field)
                
                            if hasattr(model_field,'deferred') and model_field.deferred:
                                log.debug('%s uses a deferred field converter.' % axilent_field)
                                deferred_field_converters.append((axilent_field,model_field))
                            else:
                                log.debug('%s uses an immediate field converter.' % axilent_field)
                                value = model_field.to_local_model(content_data,getattr(content_data,axilent_field))
                                fields[model_field.field] = value
                        else:
                            log.debug('Using default field converter for %s.' % axilent_field)
                            # not a field converter, just a string.  Use DefaultFieldConverter
                            default_field_converter = DefaultFieldConverter(model_field)
                            fields[model_field] = default_field_converter.to_local_model(content_data,getattr(content_data,axilent_field))
                    except AttributeError:
                        log.exception('Local model has no field %s (matched to Axilent field %s).' % (model_field,axilent_field))
                else:
                    log.info('Skipping ace field %s - not in data from ace.' % axilent_field)
            
            log.debug('Creating local model with field data %s.' % unicode(fields))
            local_model = model_class.objects.create(**fields) # create the local model with the content data
            
            if deferred_field_converters:
                for deferred_axilent_field, deferred_model_field in deferred_field_converters:
                    try:
                        deferred_model_field.to_local_model(content_data,getattr(content_data,deferred_axilent_field),local_model)
                    except AttributeError:
                        log.exception('Local model has no field %s (matched to Axilent field %s).' % (deferred_model_field.field,deferred_axilent_field))
            
                local_model.save()
            else:
                log.info('No deferred field converters for %s.' % unicode(local_model))

            local_content_type = ContentType.objects.get_for_model(local_model)
            record = self.create(local_content_type=local_content_type,
                                 local_id=local_model.pk,
                                 axilent_content_type=axilent_content_type,
                                 axilent_content_key=axilent_content_key,
                                 updated=datetime.now())
        except KeyError:
            raise ValueError('ACE content type %s cannot be found in the local registry.' % axilent_content_type)
        
        return (local_model,record)
    
    def get_or_create_model(self,axilent_content_type,axilent_content_key):
        """
        Gets the locally cached model, corresponding to the AxilentContentRecord, or,
        if there is no record, creates a new model and record from remote content
        data in ACE.
        """
        try:
            record = self.get(axilent_content_type=axilent_content_type,
                              axilent_content_key=axilent_content_key)
            return record.get_local_model()
        except AxilentContentRecord.DoesNotExist:
            local_model, record = self.create_model(axilent_content_type,axilent_content_key)
            return local_model
    
    def model_to_content_link(self,value):
        """
        Converts a local model (held as a foreign key reference) to an Axilent
        content link in <content-type>:<content-key> format.
        """
        if isinstance(value,models.Model):
            try:
                record = self.get_record(value)
                return '%s:%s' % (record.axilent_content_type,record.axilent_content_key)
            except AxilentContentRecord.DoesNotExist:
                return ''
        else:
            return value
    
    def content_link_to_model(self,value):
        """
        Converts a content link format string '<content-type>:<content-key>' to a local
        model to be used as a foreign key.
        """
        if re.match(r'^[\w\s]+:[A-Fa-f0-9]+$',value):
            ctype, ckey = value.split(':')
            record = self.get(axilent_content_type=ctype,axilent_content_key=ckey)
            return record.get_local_model()
        else:
            return value
    
    def data_for_library(self,model):
        """
        Gets a data dictionary prepared for the library.
        """
        lib_data = {}
        field_map = self.field_map(model)
        for axilent_field, model_field in field_map.items():
            try:
                lib_data[axilent_field] = self.model_to_content_link(getattr(model,model_field))
            except AttributeError:
                log.exception('Local model has no field %s (matched to Axilent field %s).' % (model_field,axilent_field))
        
        return lib_data
    
    def push_to_library(self,model):
        """
        Pushes the model to the Axilent library (assuming the library integration is active).
        
        Returns a 2-tuple of booleans indicating 1.  If the library was updated and 2. If the
        content item was created on Axilent for the first time.
        
        """
        if library_client:
            with self.lock:
                lib_data = self.data_for_library(model)
                try:
                    record = self.get_record(model)
                    # this content item already exists on Axilent - update
                    response = library_client.update_content(record.axilent_content_type,
                                                             library_project,
                                                             record.axilent_content_key,
                                                             **lib_data)
                    return (True,False)
                except AxilentContentRecord.DoesNotExist:
                    # this is new
                    local_content_type = ContentType.objects.get_for_model(model)
                    axilent_content_type = model.ACE.content_type
                    response = library_client.create_content(axilent_content_type,
                                                             library_project,
                                                             **lib_data)
                    returned_content_type, returned_key = response.split(':')
                
                    # create new record
                    self.create(local_content_type=local_content_type,
                                local_id=model.pk,
                                axilent_content_type=axilent_content_type,
                                axilent_content_key=returned_key)
            
                    return (True,True)

        else:
            return (False,False)
    
    def push_to_graphstack(self,model):
        """
        Pushes the model to the graphstack associated with the content client.  Like push_to_library
        will return a 2-tuple, indicating (1) if the graphstack was updated and (2) if it was
        created for the first time.
        """
        if content_client:
            with self.lock:
                data = self.data_for_library(model)
                try:
                    record = self.get_record(model)
                    # this content item exists in ACE, try to update
                    response = content_client.update_content(record.axilent_content_type,
                                                             record.axilent_content_key,
                                                             **data)
                    return (True,False)
                except AxilentContentRecord.DoesNotExist:
                    # new content
                    local_content_type = ContentType.objects.get_for_model(model)
                    axilent_content_type = model.ACE.content_type
                    response = content_client.create_content(axilent_content_type,
                                                             **data)
                
                    # create new record
                    self.create(local_content_type=local_content_type,
                                local_id=model.pk,
                                axilent_content_type=axilent_content_type,
                                axilent_content_key=response)
            
                    return (True,True)
        else:
            return (False,False)
    
    def search(self,model_class,query):
        """
        Searches ACE and provides model instances that match the search results.
        """
        content_type = model_class.ACE.content_type
        search_results = content_client.search(query,content_type)
        content_records = self.filter(axilent_content_type=content_type,axilent_content_key__in=[result.key for result in search_results])
        return model_class.objects.filter(pk__in=[record.local_id for record in content_records])

class AxilentContentRecord(models.Model):
    """
    Mapping for a specific Axilent content item to a model.
    """
    local_content_type = models.ForeignKey(ContentType,related_name='axilent_content_records')
    local_id = models.IntegerField()
    axilent_content_type = models.CharField(max_length=100)
    axilent_content_key = models.CharField(max_length=100)
    updated = models.DateTimeField(null=True)
    
    objects = AxilentContentRecordManager()
    
    def update_available(self):
        """
        Determines if a new content update is available from Axilent.
        """
        if content_client.latest_update(self.axilent_content_type,self.axilent_content_key):
            return True
        else:
            return False

    def get_update(self):
        """
        Gets the updated content from Axilent.
        """
        latest = content_client.latest_update(self.axilent_content_type,self.axilent_content_key)
        if self.updated and self.updated >= latest:
            return None
        else:
            return content_client.get_content(self.axilent_content_type,self.axilent_content_key)
    
    def sync_content(self,axilent_content):
        """
        Syncs the local content to the incoming axilent content (a dictionary).
        """
        from djax.content import DefaultFieldConverter
        
        local_model = self.get_local_model()
        field_map = local_model.ACE.field_map
        
        log.debug('syncing local model with Axilent content %s, using field map %s.' % (unicode(axilent_content),unicode(field_map)))
        
        # Iterate through the field map and set the local model values from the incoming Axilent content
        deferred_field_converters = []
        
        for axilent_field, model_field in field_map.items():
            if hasattr(model_field,'field'):
                # this is a field converter
                # sanity check
                if not hasattr(model_field,'to_ace') or not hasattr(model_field,'to_local_model'):
                    raise ValueError('You must define the methods to_ace and to_local_model for field converter for ace field %s.' % axilent_field)
                
                # check deferred
                if hasattr(model_field,'deferred') and model_field.deferred:
                    deferred_field_converters.append((axilent_field,model_field))
                
                # check nullable
                if (hasattr(model_field,'accepts_null') and model_field.accepts_null) or hasattr(axilent_content,axilent_field):
                    # this is a nullable field converter or a non-null value
                    value = model_field.to_local_model(axilent_content,getattr(axilent_content,axilent_field))
                    setattr(local_model,model_field.field,value)
                else:
                    log.info('Skipping non-nullable field %s - not in data from ACE.' % axilent_field)
            else:
                # just a string, use default field converter
                if hasattr(axilent_content,axilent_field):
                    default_field_converter = DefaultFieldConverter(model_field)
                    setattr(local_model,model_field,default_field_converter.to_local_model(axilent_content,getattr(axilent_content,axilent_field)))
                else:
                    log.info('Skipping default field %s - not in data from ACE.' % axilent_field)
        
        local_model.save()
        
        if deferred_field_converters:
            for deferred_axilent_field, deferred_model_field in deferred_field_converters:
                try:
                    deferred_model_field.to_local_model(axilent_content,getattr(axilent_content,deferred_axilent_field),local_model)
                except AttributeError:
                    log.exception('Local model has no field %s (matched to Axilent field %s).' % (deferred_model_field.field,deferred_axilent_field))
            
            local_model.save()
        else:
            log.info('No deferred field converters for %s.' % unicode(local_model))
            
            
        self.updated = datetime.now()
        self.save()
        
        return local_model
    
    def get_local_model(self):
        """
        Gets the local model for this record.
        """
        return self.local_content_type.model_class().objects.get(pk=self.local_id)
    
    def archive(self):
        """
        Archives the content on Axilent.
        """
        return library_client.archive_content(library_project,self.axilent_content_type,self.axilent_content_key)
    
    def live_delete(self):
        """
        Deletes the deployed version of this content from ACE.
        """
        return content_client.delete_content(self.axilent_content_type,self.axilent_content_key)
    
    def tag(self,tag_term,update_library_index=True):
        """
        Applies tag term to content.
        """
        return library_client.tag_content(library_project,
                                          self.axilent_content_type,
                                          self.axilent_content_key,
                                          tag_term,
                                          search_index=update_library_index)
    
    def detag(self,tag_term):
        """
        Removes tag from the content.
        """
        return library_client.detag_content(library_project,self.axilent_content_type,self.axilent_content_key,tag_term)
    
    def live_tag(self,tag_term):
        """
        Tags the deployed version of the content in the graphstack.
        """
        return content_client.tag_content(self.axilent_content_type,self.axilent_content_key,tag_term)
    
    def live_detag(self,tag_term):
        """
        De-tags the deployed veresion of the content in the graphstack.
        """
        return content_client.detag_content(self.axilent_content_type,self.axilent_content_key,tag_term)
    
    def reindex(self):
        """
        Re-indexes the deployed content for search.
        """
        return content_client.reindex_content(self.axilent_content_type,self.axilent_content_key)
    
    class Meta:
        unique_together = (('local_content_type','local_id'),('axilent_content_type','axilent_content_key'))

class ContentSyncLock(models.Model):
    """
    Lock for a content sync.  Indicates a sync is under way.
    """
    token = models.CharField(max_length=100)

class ProfileRecordManager(models.Manager):
    """
    Manager for the profile record.
    """
    def for_user(self,user):
        """
        Gets or creates a profile record for the user.  User may not be anonymous for this call.
        """
        if user.is_anonymous():
            raise ValueError('You cannon use an anonymous user for the ProfileRecord.objects.for_user() method.')
        
        try:
            return (self.get(user=user).profile,False)
        except ProfileRecord.DoesNotExist:
            profile = trigger_client.profile()
            record = self.create(user=user,profile=profile)
            return (record.profile,True)
    
    def for_request(self,request):
        """
        Gets or creates a profile from the request object.  Will attempt to use profile record
        registered with a logged in user, failing that will fall back to a cookie.
        
        Returns a tuple of the ProfileRecord and a boolean flag indicating if the record
        was just created.
        """
        if not request.user.is_anonymous():
            return self.for_user(request.user)
        
        profile = request.COOKIES.get('axilent-profile',None)
        profile_record = None
        if profile:
            return self.get_or_create(profile=profile)
        else:
            profile_record = self.create(profile=trigger_client.profile())
            return (profile_record,True)

class ProfileRecord(models.Model):
    """
    A record associating a Django user with an ACE profile.
    """
    user = models.ForeignKey(User,related_name='ace_profile_record',unique=True,null=True)
    profile = models.CharField(max_length=100)
    
    objects = ProfileRecordManager()
    
    def __unicode__(self):
        return self.profile

class AuthTokenManager(models.Manager):
    """
    Manager class for AuthToken.
    """
    def new_token(self,origin_domain=None):
        """
        Creates a new token.
        """
        return self.create(origin_domain=origin_domain,
                           token=uuid.uuid4().hex)

class AuthToken(models.Model):
    """
    Token for remote management of Djax install.
    """
    origin_domain = models.URLField(null=True)
    token = models.CharField(max_length=100,unique=True)
    
    objects = AuthTokenManager()
    
    def __unicode__(self):
        return self.token


# =================
# = Registry Hook =
# =================
build_registry()


"""
Registry for Djax integrations.
"""
from django.conf import settings
import inspect
from django.db.models import Model
import logging

log = logging.getLogger('djax')

content_registry = {}

class MalformedRegistry(Exception):
    """
    Indicates the Djax registry has been corrupted.
    """

def get_module(module_name):
    """
    Imports and returns the named module.
    """
    module = __import__(module_name)
    components = module_name.split('.')
    for comp in components[1:]:
        module = getattr(module,comp)
    return module

def build_registry():
    """
    Builds the registry.
    """
    from djax.content import ACEContent
    
    for app_path in settings.INSTALLED_APPS:
        if not ('djax' in app_path):
            try:
                #log.debug('Examining app %s' % app_path)
                app_module = get_module(app_path)
                if hasattr(app_module,'models'):
                    module = getattr(app_module,'models')
                    for name, attribute in inspect.getmembers(module):
                        # log.debug('content registry examining module member %s.' % name)
                        if inspect.isclass(attribute) and issubclass(attribute,Model) and issubclass(attribute,ACEContent):
                            # this is a content model, add to registry
                            try:
                                log.info('Adding model %s to content registry.' % name)
                                content_registry[attribute.ACE.content_type] = attribute
                            except AttributeError:
                                raise MalformedRegistry('All ACE content mappings must be defined with an "ACE" inner class with a "content_type" attribute.')
            except ImportError:
                log.warn('Cannot import %s.  Skipping.' % app_path)



"""
Clears out dead sync locks.
"""
from django.core.management.base import BaseCommand
from optparse import make_option
from djax.models import ContentSyncLock

class Command(BaseCommand):
    """
    Command class.
    """
    def handle(self,**options):
        """
        Handler.
        """
        for sync_lock in ContentSyncLock.objects.all():
            sync_lock.delete()



"""
Lists the registered auth tokens.
"""
from django.core.management.base import BaseCommand
from optparse import make_option
from djax.models import AuthToken

class Command(BaseCommand):
    """
    Command class.
    """
    option_list = BaseCommand.option_list + (
        make_option('--host',dest='host',default=None),
    )
    
    def handle(request,*args,**options):
        """
        Handler method.
        """
        host = options['host']
        tokens = AuthToken.objects.all()
        if host:
            tokens = tokens.filter(origin_domain=host)
        
        for token in tokens:
            if token.origin_domain:
                print token.token,token.origin_domain
            else:
                print token.token



"""
Utilities.
"""
import unicodedata
import re

def slugify(value):
    """
    Slugs the string.
    """
    value = unicodedata.normalize('NFKD', unicode(value)).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^\w\s-]', '', value).strip().lower()
    return re.sub('[-\s]+', '-', value)



"""
Content API.
"""
from pax.exceptions import PaxException
from pax.util import slugify
from dateutil import parser

class ContentImage(object):
    """
    Wrapper for content item from Axilent.
    """
    def __init__(self,data):        
        content = None
        
        if 'endorsement' in data:
            # this is a policy result
            self.endorsement = data['endorsement']
            content = data['content']
        else:
            self.endorsement = 0
            content = data
        
        self.content_type = content['content_type']
        self.key = content['key']
        self.data = content['data']
    
    def __getattr__(self,attribute):
        """
        Fallback to data access
        """
        try:
            return self.data[attribute]
        except KeyError:
            raise AttributeError
    
    def __unicode__(self):
        """
        Show the data.
        """
        return unicode(self.data)

class ChannelResult(object):
    """
    Wrapper for content channel result.
    """
    def __init__(self,data):
        items = None
        if hasattr(data,'keys'):
            # this is a policy result
            items = data['default']
            self.flavor = data['meta']['flavor']
            self.channel = data['meta']['channel']
        else:
            items = data
            self.flavor = None
            self.channel = None
        
        self.items = [ContentImage(item_data) for item_data in items]
    
    def __iter__(self):
        return iter(self.items)

class ContentClient(object):
    """
    Content client.
    """
    def __init__(self,axilent_connection):
        self.content_resource = axilent_connection.resource_client('axilent.content','content')
        self.api = axilent_connection.http_client('axilent.content')
    
    def get_content(self,content_type,key):
        """
        Gets the specified content item.
        """
        data = self.content_resource.get(params={'content_type_slug':slugify(content_type),'content_key':key})
        return ContentImage(data)
    
    def create_content(self,content_type,**content):
        """
        Creates a new content item.
        """
        response = self.content_resource.post(data={'content_type_slug':slugify(content_type),
                                                    'content':content})
        return response['created']
    
    def update_content(self,content_type,key,**content):
        """
        Updates content.
        """
        response = self.content_resource.put(data={'content_type_slug':slugify(content_type),
                                                   'content_key':key,
                                                   'content':content})
        return response['updated']
    
    def delete_content(self,content_type,key):
        """
        Deletes the specified content.
        """
        response = self.content_resource.delete(params={'content_type_slug':slugify(content_type),
                                                        'content_key':key})
        return response['deleted']
    
    def tag_content(self,content_type,key,tag):
        """
        Tags the specified content.
        """
        response = self.api.tagcontent(content_type_slug=slugify(content_type),
                                       content_key=key,
                                       tag=tag)
        return response['tagged']
    
    def detag_content(self,content_type,key,tag):
        """
        De-tags the content.
        """
        response = self.api.detagcontent(content_type_slug=slugify(content_type),
                                         content_key=key,
                                         tag=tag)
        return response['detagged']
    
    def reindex_content(self,content_type,key):
        """
        Forces content to re-index for search.
        """
        response = self.api.reindexcontent(content_type_slug=slugify(content_type),
                                           content_key=key)
        return response['reindexed']
        
    def channel_group(self,group,profile=None,basekey=None,limit=None,flavor=None):
        """
        Gets content from a Content Channel Group.
        """
        response = self.api.contentchannelgroup(group=group,profile=profile,basekey=basekey,limit=limit,flavor=flavor)
        return ChannelResult(response)
    
    def search(self,query,*content_types):
        """
        Searches for content.
        """
        content_type_list = ','.join([slugify(ctype) for ctype in content_types])
        response = self.api.search(query=query,content_types=content_type_list)
        return ChannelResult(response)
    
    def channel(self,channel_name,profile=None,basekey=None,limit=None,flavor=None):
        """
        Gets content from a Content Channel.
        """
        response = self.api.contentchannel(channel=channel_name,profile=profile,basekey=basekey,limit=limit,flavor=flavor)
        return ChannelResult(response)
    
    def content_keys(self,content_type):
        """
        Gets a list of keys for content of the specified type.
        """
        return self.api.getcontentkeys(content_type_slug=slugify(content_type))

    def get_content_by_unique_field(self,content_type,field_name,field_value):
        """
        Gets a content item matching the specified field value.
        """
        data = self.api.getcontentbyuniquefield(content_type=slugify(content_type),field_name=field_name,field_value=field_value)
        return ContentImage(data)
    
    def latest_update(self,content_type,content_key):
        """
        Gets the date of the latest update.
        """
        response = self.api.latestupdate(content_type_slug=slugify(content_type),content_key=content_key)
        return parser.parse(response['updated']) if response['updated'] else None



"""
Main client setup for Pax.
"""
from sharrock.client import HttpClient, ResourceClient, ServiceException
import logging

log = logging.getLogger('axilent-pax')

class AxilentConnection(object):
    """
    A connection with Axilent.
    """
    def __init__(self,apikey,api_version='beta3',endpoint='https://www.axilent.net'):
        self.apikey = apikey
        self.version = api_version
        self.endpoint = endpoint
    
    def http_client(self,app):
        """
        Gets an HTTP client for the specified app.
        """
        return HttpClient('%s/api' % self.endpoint,app,self.version,auth_user=self.apikey)
    
    def resource_client(self,app,resource):
        """
        Gets a resource client for the specified app.
        """
        return ResourceClient('%s/api/resource' % self.endpoint,app,self.version,resource,auth_user=self.apikey)



"""
Network ops for Djax.  Uses the Sharrock client.
"""
from django.conf import settings
from django.template.defaultfilters import slugify
import logging
from dateutil import parser

from pax.client import AxilentConnection
from pax.content import ContentClient
from pax.library import LibraryClient
from pax.triggers import TriggerClient

log = logging.getLogger('djax')

# ============
# = Settings =
# ============
_endpoint = 'https://www.axilent.net'
if hasattr(settings,'AXILENT_ENDPOINT') and settings.AXILENT_ENDPOINT:
    _endpoint = settings.AXILENT_ENDPOINT

_api_version = 'beta3'
if hasattr(settings,'AXILENT_API_VERSION') and settings.AXILENT_API_VERSION:
    _api_version = settings.AXILENT_API_VERSION

if not hasattr(settings,'AXILENT_API_KEY') or not settings.AXILENT_API_KEY:
    raise ValueError('You must set the AXILENT_API_KEY in Django settings.')

_api_key = settings.AXILENT_API_KEY

_library_api_key = settings.AXILENT_LIBRARY_API_KEY if hasattr(settings,'AXILENT_LIBRARY_API_KEY') else None

# ===========
# = Clients =
# ===========
cx = AxilentConnection(_api_key,_api_version,_endpoint)
library_cx = AxilentConnection(_library_api_key,_api_version,_endpoint) if _library_api_key else None

content_client = ContentClient(cx)
trigger_client = TriggerClient(cx)
library_client = LibraryClient(library_cx) if library_cx else None
library_project = settings.AXILENT_LIBRARY_PROJECT if hasattr(settings,'AXILENT_LIBRARY_PROJECT') else None


"""
Local API for messaging plugin.
"""

class MessagingClient(object):
    """
    Client for messaging system.
    """
    def __init__(self,axilent_connection):
        """
        Constructor.
        """
        self.message_resource = axilent_connection.resource_client('axilent.plugins.messaging','message')
        self.recipient_resource = axilent_connection.resource_client('axilent.plugins.messaging','recipient')
        self.received_message_resource = axilent_connection.resource_client('axilent.plugins.messaging','receivedmessage')
        self.api = axilent_connection.http_client('axilent.plugins.messaging')
    
    def create_message(self,bus,message_type,sender,content):
        """
        Creates a new message.
        """
        response = self.message_resource.post(data={'bus':bus,
                                                    'message_type':message_type,
                                                    'sender':sender,
                                                    'content':content})
        return response['created']
    
    def create_recipient(self,bus,recipient_type,content):
        """
        Creates a new recipient.
        """
        response = self.recipient_resource.post(data={'bus':bus,
                                                      'recipient_type':recipient_type,
                                                      'content':content})
        return response['created']
    
    def update_recipient(self,key,content):
        """
        Updates an existing recipient.
        """
        response = self.recipient_resource.put(data={'key':key,
                                                     'content':content})
        
        return response['updated']
    
    def get_recipient(self,key):
        """
        Gets the specified recipient.
        """
        response = self.recipient_resource.get(params={'key':key})
        return response['content']
    
    def delete_recipient(self,key):
        """
        Deletes the specified recipient.
        """
        response = self.recipient_resource.delete(params={'key':key})
        return response['deleted']
    
    
    def publish_message(self,key,topic):
        """
        Publishes the specified message on the topic.
        """
        response = self.api.publishmessage(key=key,topic=topic)
        return response['sent']
    
    def send_message(self,message_key,recipient_key):
        """
        Point to point sending of the message to the specified recipient.
        """
        response = self.api.sendmessage(message_key=message_key,recipient_key=recipient_key)
        return response['sent']
    
    def inbox(self,recipient_key,unread_only=True):
        """
        Gets the inbox messages for the recipient.
        """
        unread_string = 'True' if unread_only else 'False'
        response = self.api.inbox(recipient_key=recipient_key,unread_only=unread_string)
        return response['received-messages']
    
    def get_received_message(self,recipient_key,message_key):
        """
        Gets the received message for the recipient.
        """
        return self.received_message_resource.get(params={'recipient_key':recipient_key,'message_key':message_key})
        
    def update_received_message(self,recipient_key,message_key,read=True):
        """
        Updates the received message (changing the 'read' status).
        """
        read_string = 'True' if read else 'False'
        response = self.received_message_resource.put(params={'recipient_key':recipient_key,'message_key':message_key,'read':read_string})
        return bool(response['message-read'])
    
    def delete_received_message(self,recipient_key,message_key):
        """
        Deletes the received message.
        """
        response = self.received_message_resource.delete(params={'recipient_key':recipient_key,'message_key':message_key})
        return response['message-deleted']
    
    def subscribe(self,recipient_key,topic):
        """
        Subscribes the recipient to the topic.
        """
        response = self.api.subscribe(recipient_key=recipient_key,topic=topic)
        return response['topic-subscribed']
    
    def unsubscribe(self,recipient_key,topic):
        """
        Unsubscribes the recipient from the topic.
        """
        response = self.api.unsubscribe(recipient_key=recipient_key,topic=topic)
        return response['topic-subscribed']

        


#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError, err:
    from distutils.core import setup, find_packages
    
from djax import VERSION

setup(
    name='Djax',
    version='.'.join(map(str,VERSION)),
    description='Integrates Django projects with Axilent.',
    packages=find_packages(exclude=['example']),
    license='BSD',
    author='Loren Davie',
    author_email='code@axilent.com',
    url='http://github.com/Axilent/Djax',
    install_requires=['sharrock-client','Django','python-dateutil'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)


"""
Triggers API.
"""
from pax.exceptions import PaxException
import logging

log = logging.getLogger('djax')

class TriggerClient(object):
    """
    Client for Triggers API.
    """
    def __init__(self,axilent_connection):
        self.api = axilent_connection.http_client('axilent.triggers')
    
    def trigger(self,category,action,profile=None,variables={},environment={},identity={}):
        """
        Sends a trigger to Axilent.
        """
        self.api.trigger(data={'category':category,
                               'action':action,
                               'profile':profile,
                               'variables':variables,
                               'environment':environment,
                               'identity':identity},
                               local_param_check=False)


    def profile(self):
        """
        Gets a profile to use with triggers and content channels.
        """
        return self.api.profile()['profile']



"""
Synchronizes the local model with Axilent.
"""
from django.core.management.base import BaseCommand
from optparse import make_option
from djax.content import sync_content

class Command(BaseCommand):
    """
    The command class.
    """
    option_list = BaseCommand.option_list + (
        make_option('--content-type',
                    dest='content_type',
                    default=None,
                    help='The ACE content type to sync.  If not specified then all the ACE content will by synced.'),
    )
    
    def handle(self,*args,**options):
        """
        Handler method.
        """
        print 'Syncing local models with ACE'
        content_type = options.get('content_type',None)
        result = None
        if content_type:
            result = sync_content(content_type_to_sync=content_type)
        else:
            result = sync_content()
        if result:
            print 'Content model has been synced with ACE'
        else:
            print '''Local model is sync-locked, suggesting a concurrent sync is underway.  
                     If you think this is an error, clear the lock by running "manage.py 
                     clear_content_sync_locks".'''
